@@ -85,6 +85,9 @@ static struct glvnd_list displayList = { &displayList, &displayList };
 static glvnd_mutex_t displayListLock = GLVND_MUTEX_INITIALIZER;
 static EGLint failNextMakeCurrentError = EGL_NONE;
 
+static glvnd_mutex_t contextListLock = GLVND_MUTEX_INITIALIZER;
+static struct glvnd_list contextList = { &contextList, &contextList };
+
 static EGLDEBUGPROCKHR debugCallbackFunc = NULL;
 static EGLBoolean debugCallbackEnabled = EGL_TRUE;
 
@@ -112,9 +115,20 @@ static DummyThreadState *GetThreadState(void)
     return thr;
 }
 
+void DestroyThreadState(DummyThreadState *thr)
+{
+    if (thr != NULL)
+    {
+        __glvndPthreadFuncs.mutex_lock(&threadStateLock);
+        glvnd_list_del(&thr->entry);
+        __glvndPthreadFuncs.mutex_unlock(&threadStateLock);
+        free(thr);
+    }
+}
+
 static void OnThreadTerminate(void *ptr)
 {
-    free(ptr);
+    DestroyThreadState((DummyThreadState *) ptr);
 }
 
 static void CommonEntrypoint(void)
@@ -326,6 +340,10 @@ static EGLContext EGLAPIENTRY dummy_eglCreateContext(EGLDisplay dpy,
     dctx = (DummyEGLContext *) calloc(1, sizeof(DummyEGLContext));
     dctx->vendorName = DUMMY_VENDOR_NAME;
 
+    __glvndPthreadFuncs.mutex_lock(&contextListLock);
+    glvnd_list_append(&dctx->entry, &contextList);
+    __glvndPthreadFuncs.mutex_unlock(&contextListLock);
+
     return (EGLContext) dctx;
 }
 
@@ -336,6 +354,9 @@ static EGLBoolean EGLAPIENTRY dummy_eglDestroyContext(EGLDisplay dpy, EGLContext
 
     if (ctx != EGL_NO_CONTEXT) {
         DummyEGLContext *dctx = (DummyEGLContext *) ctx;
+        __glvndPthreadFuncs.mutex_lock(&contextListLock);
+        glvnd_list_del(&dctx->entry);
+        __glvndPthreadFuncs.mutex_unlock(&contextListLock);
         free(dctx);
     }
     return EGL_TRUE;
@@ -504,7 +525,7 @@ static EGLBoolean EGLAPIENTRY dummy_eglReleaseThread(void)
         __glvndPthreadFuncs.getspecific(threadStateKey);
     if (thr != NULL) {
         __glvndPthreadFuncs.setspecific(threadStateKey, NULL);
-        free(thr);
+        DestroyThreadState(thr);
     }
     return EGL_TRUE;
 }
@@ -957,6 +978,14 @@ void _fini(void)
         return;
     }
 
+    __glvndPthreadFuncs.mutex_lock(&contextListLock);
+    while (!glvnd_list_is_empty(&contextList)) {
+        DummyEGLContext *ctx = glvnd_list_first_entry(&contextList, DummyEGLContext, entry);
+        glvnd_list_del(&ctx->entry);
+        free(ctx);
+    }
+    __glvndPthreadFuncs.mutex_unlock(&contextListLock);
+
     while (!glvnd_list_is_empty(&displayList)) {
         DummyEGLDisplay *disp = glvnd_list_first_entry(
                 &displayList, DummyEGLDisplay, entry);
